@@ -3,6 +3,9 @@ from .grid import GridSearch
 from .smbo import SMBOSearch
 from .nevergrad import NevergradSearch
 from .coordinate import CoordinateSearch
+from .halving import HalvingSearch
+from .hyperband import HyperbandSearch
+from .pbt import PBTSearch
 
 
 __all__ = [
@@ -12,4 +15,7 @@ __all__ = [
     "SMBOSearch",
     "NevergradSearch",
     "CoordinateSearch",
+    "HalvingSearch",
+    "HyperbandSearch",
+    "PBTSearch",
 ]
