@@ -1,6 +1,6 @@
 from ..strategy import Strategy
 from ..spaces import NevergradSpace
-from typing import Union
+from typing import Union, List
 import nevergrad as ng
 
 
@@ -22,6 +22,7 @@ class NevergradSearch(Strategy):
         seed_id: int = 42,
         verbose: bool = False,
     ):
+        self.search_name = "Nevergrad"
         Strategy.__init__(
             self,
             real,
@@ -37,7 +38,6 @@ class NevergradSearch(Strategy):
         )
         self.space = NevergradSpace(real, integer, categorical)
         self.init_optimizer()
-        self.search_name = "Nevergrad Wrapper Search"
 
         # Add start-up message printing the search space
         if self.verbose:
@@ -66,7 +66,12 @@ class NevergradSearch(Strategy):
         param_batch = [params.value[1] for params in last_batch_params]
         return param_batch
 
-    def tell_search(self, batch_proposals, perf_measures):
+    def tell_search(
+        self,
+        batch_proposals: list,
+        perf_measures: list,
+        ckpt_paths: Union[List[str], None] = None,
+    ):
         """Perform post-iteration clean-up by updating surrogate model."""
         for i, prop in enumerate(batch_proposals):
             # Need to update hyperoptimizer with ng Instrumentation candidate
@@ -86,6 +91,40 @@ class NevergradSearch(Strategy):
                 else:
                     self.hyper_optimizer.tell(x, [-1 * p for p in perf_measures[i]])
 
+    def update_search(self):
+        """Refine search space boundaries after set of search iterations."""
+        if self.refine_after is not None:
+            # Check whether there are still refinements open
+            # And whether we have already passed last refinement point
+            if len(self.refine_after) > self.refine_counter:
+                exact = self.eval_counter == self.refine_after[self.refine_counter]
+                skip = (
+                    self.eval_counter > self.refine_after[self.refine_counter]
+                    and self.last_refined != self.refine_after[self.refine_counter]
+                )
+                if exact or skip:
+                    self.refine(self.refine_top_k)
+                    self.last_refined = self.refine_after[self.refine_counter]
+                    self.refine_counter += 1
+
+    def setup_search(self):
+        """Initialize search settings at startup."""
+        # Set up search space refinement - random, SMBO, nevergrad
+        if self.search_config is not None:
+            if "refine_top_k" in self.search_config.keys():
+                self.refine_counter = 0
+                assert self.search_config["refine_top_k"] > 1
+                self.refine_after = self.search_config["refine_after"]
+                # Make sure that refine iteration is list
+                if type(self.refine_after) == int:
+                    self.refine_after = [self.refine_after]
+                self.refine_top_k = self.search_config["refine_top_k"]
+                self.last_refined = 0
+            else:
+                self.refine_after = None
+        else:
+            self.refine_after = None
+
     def refine_space(self, real, integer, categorical):
         """Update the nevergrad search space."""
         self.space.update(real, integer, categorical)
@@ -96,10 +135,10 @@ class NevergradSearch(Strategy):
 
     def get_pareto_front(self):
         """Get the pareto-front of the optimizer."""
-        pareto_configs, pareto_evals = [], []
+        pareto_configs, pareto_evals, pareto_ckpt = [], [], None
         for param in sorted(
             self.hyper_optimizer.pareto_front(), key=lambda p: p.losses[0]
         ):
             pareto_configs.append(param.value[1])
             pareto_evals.append(param.losses)
-        return pareto_configs, pareto_evals
+        return pareto_configs, pareto_evals, pareto_ckpt
